'''
## This file is autogenerated from macrosynergy setup.py.

It contains version information for the macrosynergy package.

short_version: 0.1.26

version: 0.1.26

full_version: 0.1.26

git_revision: 2d430bf809a99b726a9e691bc4710522d99a9006

release: True

'''
short_version: str = '0.1.26'
version: str = '0.1.26'
full_version: str = '0.1.26'
git_revision: str = '2d430bf809a99b726a9e691bc4710522d99a9006'
release: bool = True
if not release:
    version = full_version
